# coding: utf-8
#
# Copy from (nose)[http://nose.readthedocs.io/en/latest/]:
# - nose.tools.with_setup

"""Tools to help writing tests
"""


def with_setup(setup=None, teardown=None):
    """Decorator to add setup and/or teardown methods to a test function::

      @with_setup(setup, teardown)
      def test_something():
          " ... "

    Note that `with_setup` is useful *only* for test functions, not for test
    methods or inside of TestCase subclasses.
    """
    def decorate(func, setup=setup, teardown=teardown):
        if setup:
            if hasattr(func, 'setup'):
                _old_s = func.setup

                def _s():
                    setup()
                    _old_s()
                func.setup = _s
            else:
                func.setup = setup
        if teardown:
            if hasattr(func, 'teardown'):
                _old_t = func.teardown

                def _t():
                    _old_t()
                    teardown()
                func.teardown = _t
            else:
                func.teardown = teardown
        return func
    return decorate



# coding: utf-8

import os
import sys
import imp
import logging

lg = logging.getLogger('deptest.loader')


def load_module_from_path(filepath):
    # add path
    dirpath = os.path.dirname(filepath)
    sys.path.insert(0, dirpath)

    module_name = os.path.basename(filepath).split('.')[0]
    lg.debug('load module %s from %s, at %s', module_name, filepath, os.getcwd())
    module = imp.load_source(module_name, filepath)

    # remove path
    del sys.path[0]

    return module



# coding: utf-8

import sys
import curses
import logging
import threading
from .utils import anyp, unicode_type, to_unicode

lg = logging.getLogger('deptest.log')


def setup_log_handler(log_handler, clear=False):
    # setup our handler with root logger
    root_logger = logging.getLogger()
    if clear:
        if hasattr(root_logger, "handlers"):
            for handler in root_logger.handlers:
                root_logger.removeHandler(handler)
        for logger in logging.Logger.manager.loggerDict.values():
            if hasattr(logger, "handlers"):
                for handler in logger.handlers:
                    logger.removeHandler(handler)
    # make sure there isn't one already
    # you can't simply use "if log_handler not in root_logger.handlers"
    # since at least in unit tests this doesn't work --
    # LogCapture() is instantiated for each test case while root_logger
    # is module global
    # so we always add new MyMemoryHandler instance
    for handler in root_logger.handlers[:]:
        if isinstance(handler, MyMemoryHandler):
            root_logger.handlers.remove(handler)
    root_logger.addHandler(log_handler)
    lg.debug('root logger handlers: %s', root_logger.handlers)
    # to make sure everything gets captured
    loglevel = "NOTSET"
    root_logger.setLevel(getattr(logging, loglevel))


def set_logger(name,
               level=logging.INFO,
               fmt='%(name)s %(levelname)s %(message)s',
               propagate=1):
    """
    This function will clear the previous handlers and set only one handler,
    which will only be StreamHandler for the logger.

    This function is designed to be able to called multiple times in a context.

    Note that if a logger has no handlers, it will be added a handler automatically when it is used.
    """
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = propagate

    handler = None
    for h in logger.handlers:
        if isinstance(h, logging.StreamHandler):
            # use existing instead of clean and create
            handler = h
            break
    if not handler:
        handler = logging.StreamHandler()
        logger.addHandler(handler)

    handler.setFormatter(logging.Formatter(fmt=fmt))


class MyMemoryHandler(logging.Handler):
    def __init__(self, logformat, logdatefmt=None, filters=None):
        logging.Handler.__init__(self)
        fmt = logging.Formatter(logformat, logdatefmt)
        self.setFormatter(fmt)
        if filters is None:
            filters = ['-deptest']
        self.filterset = FilterSet(filters)
        self.buffer = []

    def emit(self, record):
        self.buffer.append(self.format(record))

    def flush(self):
        pass

    def truncate(self):
        self.buffer = []

    def filter(self, record):
        if self.filterset.allow(record.name):
            return logging.Handler.filter(self, record)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['lock']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.lock = threading.RLock()


class FilterSet(object):
    def __init__(self, filter_components):
        self.inclusive, self.exclusive = self._partition(filter_components)

    # @staticmethod
    def _partition(components):
        inclusive, exclusive = [], []
        for component in components:
            if component.startswith('-'):
                exclusive.append(component[1:])
            else:
                inclusive.append(component)
        return inclusive, exclusive
    _partition = staticmethod(_partition)

    def allow(self, record):
        """returns whether this record should be printed"""
        if not self:
            # nothing to filter
            return True
        return self._allow(record) and not self._deny(record)

    # @staticmethod
    def _any_match(matchers, record):
        """return the bool of whether `record` starts with
        any item in `matchers`"""
        def record_matches_key(key):
            return record == key or record.startswith(key + '.')
        return anyp(bool, map(record_matches_key, matchers))
    _any_match = staticmethod(_any_match)

    def _allow(self, record):
        if not self.inclusive:
            return True
        return self._any_match(self.inclusive, record)

    def _deny(self, record):
        if not self.exclusive:
            return False
        return self._any_match(self.exclusive, record)


class Color(object):
    DEFAULT_COLORS = {
        'blue': 4,
        'green': 2,
        'yellow': 3,
        'red': 1,
    }

    def __init__(self, colors=DEFAULT_COLORS):
        enable = False
        if curses and hasattr(sys.stderr, 'isatty') and sys.stderr.isatty():
            try:
                curses.setupterm()
                if curses.tigetnum("colors") > 0:
                    enable = True
            except Exception:
                pass

        self._colors = {}

        if enable:
            fg_color = (curses.tigetstr("setaf") or
                        curses.tigetstr("setf") or "")
            if (3, 0) < sys.version_info < (3, 2, 3):
                fg_color = unicode_type(fg_color, "ascii")

            for name, code in colors.items():
                self._colors[name] = unicode_type(curses.tparm(fg_color, code), "ascii")
            self._normal = unicode_type(curses.tigetstr("sgr0"), "ascii")
        else:
            self._normal = ''

        self.enable = enable

    def dye(self, name, s):
        if name in self._colors:
            start = self._colors[name]
        else:
            start = self._normal
        end = self._normal
        ctx = {'start': start, 'end': end, 'content': to_unicode(s)}
        us = u'%(start)s%(content)s%(end)s' % ctx
        return us.encode('utf8')

color = Color()



# coding: utf-8
# flake8: noqa

"""Assert functions to make tests easier

these functions are exported from methods of unitest.TestCase class,
the original camel case name is converted to underscore style,
below is a complete list of function names compare to unittest method names:

almost_equal                  assertAlmostEqual
dict_contains_subset          assertDictContainsSubset
dict_equal                    assertDictEqual
equal                         assertEqual
false                         assertFalse
greater                       assertGreater
greater_equal                 assertGreaterEqual
in_                           assertIn
is_                           assertIs
is_instance                   assertIsInstance
is_none                       assertIsNone
is_not                        assertIsNot
is_not_none                   assertIsNotNone
items_equal                   assertItemsEqual
less                          assertLess
less_equal                    assertLessEqual
list_equal                    assertListEqual
multi_line_equal              assertMultiLineEqual
not_almost_equal              assertNotAlmostEqual
not_equal                     assertNotEqual
not_in                        assertNotIn
not_is_instance               assertNotIsInstance
not_regexp_matches            assertNotRegexpMatches
raises                        assertRaises
raises_regexp                 assertRaisesRegexp
regexp_matches                assertRegexpMatches
sequence_equal                assertSequenceEqual
set_equal                     assertSetEqual
true                          assertTrue
tuple_equal                   assertTupleEqual
"""

import unittest


class MethodMaker(object):
    def __init__(self):
        class Dummy(unittest.TestCase):
            def nop():
                pass

        self.t = Dummy('nop')

    def make_func(self, func_name, method_name):
        method = getattr(self.t, method_name)

        def func(*args, **kwargs):
            try:
                return method(*args, **kwargs)
            except AssertionError as e:
                raise e

        func.__name__ = func_name
        func.__doc__ = method.__doc__

        return func


_m = MethodMaker


almost_equal         = _m.make_func('almost_equal',         'assertAlmostEqual')
dict_contains_subset = _m.make_func('dict_contains_subset', 'assertDictContainsSubset')
dict_equal           = _m.make_func('dict_equal',           'assertDictEqual')
equal                = _m.make_func('equal',                'assertEqual')
false                = _m.make_func('false',                'assertFalse')
greater              = _m.make_func('greater',              'assertGreater')
greater_equal        = _m.make_func('greater_equal',        'assertGreaterEqual')
in_                  = _m.make_func('in_',                  'assertIn')
is_                  = _m.make_func('is_',                  'assertIs')
is_instance          = _m.make_func('is_instance',          'assertIsInstance')
is_none              = _m.make_func('is_none',              'assertIsNone')
is_not               = _m.make_func('is_not',               'assertIsNot')
is_not_none          = _m.make_func('is_not_none',          'assertIsNotNone')
items_equal          = _m.make_func('items_equal',          'assertItemsEqual')
less                 = _m.make_func('less',                 'assertLess')
less_equal           = _m.make_func('less_equal',           'assertLessEqual')
list_equal           = _m.make_func('list_equal',           'assertListEqual')
multi_line_equal     = _m.make_func('multi_line_equal',     'assertMultiLineEqual')
not_almost_equal     = _m.make_func('not_almost_equal',     'assertNotAlmostEqual')
not_equal            = _m.make_func('not_equal',            'assertNotEqual')
not_in               = _m.make_func('not_in',               'assertNotIn')
not_is_instance      = _m.make_func('not_is_instance',      'assertNotIsInstance')
not_regexp_matches   = _m.make_func('not_regexp_matches',   'assertNotRegexpMatches')
raises               = _m.make_func('raises',               'assertRaises')
raises_regexp        = _m.make_func('raises_regexp',        'assertRaisesRegexp')
regexp_matches       = _m.make_func('regexp_matches',       'assertRegexpMatches')
sequence_equal       = _m.make_func('sequence_equal',       'assertSequenceEqual')
set_equal            = _m.make_func('set_equal',            'assertSetEqual')
true                 = _m.make_func('true',                 'assertTrue')
tuple_equal          = _m.make_func('tuple_equal',          'assertTupleEqual')


del _m
del MethodMaker



# coding: utf-8


class Config(object):
    _internal_keys = ['keys', 'args_keys', 'parser']

    def __init__(self):
        self.keys = []
        self.args_keys = []
        self.parser = None

    def register_parser(self, parser):
        self.parser = parser

    def define(self, key, typ):
        self.keys.append(key)

        if typ == 'args':
            self.args_keys.append(key)
            setattr(self, key, NotImplementedError)
        else:
            setattr(self, key, None)

    def __setattr__(self, key, value):
        if key not in self._internal_keys and key not in self.keys:
            raise ValueError('%s is not defined in config' % key)
        super(Config, self).__setattr__(key, value)

    def parse_args(self):
        args = self.parser.parse_args()

        for i in self.args_keys:
            setattr(self, i, getattr(args, i))

    def __str__(self):
        return '<Config: {}>'.format(
            ','.join('{}={}'.format(i, getattr(self, i)) for i in self.keys)
        )



#!/usr/bin/env python
# coding=utf-8

from setuptools import setup


# Use semantic versioning: MAJOR.MINOR.PATCH
version = '0.1.6'


def get_long_description():
    try:
        with open('README.md', 'r') as f:
            return f.read()
    except IOError:
        return ''


setup(
    # license='License :: OSI Approved :: MIT License',
    name='deptest',
    version=version,
    author='reorx',
    author_email='novoreorx@gmail.com',
    description='dependent testing framework',
    url='https://github.com/reorx/deptest',
    long_description=get_long_description(),
    packages=[
        'deptest',
    ],
    # Or use (make sure find_packages is imported from setuptools):
    # packages=find_packages()
    # Or if it's a single file package
    # install_requires=get_requires(),
    # package_data={}
    entry_points={
        'console_scripts': [
            'deptest = deptest.core:main'
        ]
    }
)



# coding: utf-8

import os
import re
import sys
import logging
import argparse
import traceback
from StringIO import StringIO
from collections import defaultdict

from .loader import load_module_from_path
from .utils import ln, hr, safe_str, merge_list, ObjectDict
from .log import setup_log_handler, MyMemoryHandler, set_logger, color
from .config_object import Config


lg = logging.getLogger('deptest')

config = Config()

STATUS_NAMES = ['OK', 'FAILED', 'UNMET']

STATUS_COLORS = {
    'OK': 'green',
    'FAILED': 'red',
    'UNMET': 'yellow',
}

COLORED_STATUSES = {i: color.dye(STATUS_COLORS[i], i) for i in STATUS_NAMES}


class ModuleRunner(object):
    entry_pattern = re.compile(r'^test_\w+$')
    module_setup_pattern = re.compile(r'^global_setup$')
    module_teardown_pattern = re.compile(r'^global_teardown$')

    def __init__(self, path):
        # parse path
        if ':' in path:
            module_path, entry_name = tuple(path.split(':'))
        else:
            module_path, entry_name = path, None
        lg.debug('module_path: {}, entry_name: {}'.format(module_path, entry_name))
        module = self.load_module(module_path)

        entries = []
        entries_dict = {}
        module_setup = None
        module_teardown = None

        for name in dir(module):
            attr = getattr(module, name)

            # get entry
            if self.entry_pattern.match(name):
                lg.debug('match entry %s', name)

                # add essential attributes for entry
                if not hasattr(attr, 'dependencies'):
                    attr.dependencies = []

                entries.append(attr)
                entries_dict[attr.__name__] = attr
                continue

            # get setup
            if self.module_setup_pattern.match(name):
                lg.debug('match module_setup %s', name)
                module_setup = attr
                continue

            # get teardown
            if self.module_teardown_pattern.match(name):
                lg.debug('match module_teardown %s', name)
                module_teardown = attr
                continue

        self.entries = entries
        self.entries_dict = entries_dict
        self._traverse_entries()

        self.module_setup = module_setup
        self.module_teardown = module_teardown
        self.module = module
        self.module_path = module_path

        # filter entries
        if entry_name:
            _entries_to_run = set([self.entries_dict[entry_name]])
            # make sure each entry in pendings got their deps
            for i in list(_entries_to_run):
                deps = traverse_entry_dependencies(i, self.entries_dict)
                _entries_to_run |= set(deps)
            self.entries_to_run = list(_entries_to_run)
        else:
            self.entries_to_run = self.entries
            # TODO tag support

        lg.debug('entries to run: %s', [i.__name__ for i in self.entries_to_run])

    def load_module(self, path):
        module = load_module_from_path(path)
        lg.debug('ModuleRunner init: %s', module)
        return module

    def _traverse_entries(self):
        lg.debug('entries_dict: %s', self.entries_dict)
        for entry in self.entries:
            deps = traverse_entry_dependencies(entry, self.entries_dict)
            lg.debug('entry {} depend on {}'.format(entry.__name__, deps))

    def dispatch(self):
        lg.debug('ModuleRunner dispatch')
        states = defaultdict(get_state)
        self.states = states
        self._dispatch(self.entries_to_run, states)

    def _dispatch(self, entries, states):
        lg.debug('_dispatch')
        pendings = []

        for entry in entries:
            deps = traverse_entry_dependencies(entry, self.entries_dict)
            state = states[entry]
            if deps:
                if should_unmet(deps, states):
                    set_state(state, 'unmet', True)
                elif should_pending(deps, states):
                    lg.debug('%s PENDING', entry)
                    pendings.append(entry)
                    continue

            self.run_entry(entry, states)

        if pendings:
            lg.debug('pendings: %s', [i.__name__ for i in pendings])
            self._dispatch(pendings, states)

    def run_entry(self, entry, states):
        lg.debug('run entry %s', entry)
        state = states[entry]
        entry_runner = EntryRunner(entry, state, self)
        entry_runner.run()

    def __str__(self):
        return '<ModuleRunner: {}>'.format(self.module.__name__)


class EntryRunner(object):
    def __init__(self, entry, state, module_runner):
        self.entry = entry
        self.state = state
        self.module_runner = module_runner
        self.stdout = []
        self._buf = None
        self.output = None

        # config
        self.clear = False

    def run(self):
        entry = self.entry
        state = self.state

        if state['unmet']:
            lg.debug('%s UNMET, skip run', entry.__name__)
            self.log_state()
            return

        self.before()

        try:
            args = []
            for i in entry.dependencies:
                dep = self.module_runner.entries_dict[i['name']]
                dep_state = self.module_runner.states[dep]
                with_return = i['with_return']
                if with_return:
                    args.append(dep_state['return_value'])
                #lg.info('dep %s %s', dep, dep_state)
            state['return_value'] = entry(*args)
        except:
            state['traceback'] = traceback.format_exc()
            state['ok'] = False
        else:
            state['ok'] = True
        finally:
            state['executed'] = True

        self.after()

        # log state
        self.log_state()

    def before(self):
        if not config.nocapture:
            # capture_stdout
            self.capture_stdout()

        if not config.nologcapture:
            # capture_logging
            self.capture_logging()

    def after(self):
        state = self.state

        if not config.nocapture:
            # get output
            state['captured_stdout'] = self._get_buffer()

            # restore_stdout
            self.restore_stdout()

        if not config.nologcapture:
            # get logging
            state['captured_logging'] = self._get_logging()

            # restore_logging
            self.restore_logging()

    def capture_stdout(self):
        self.stdout.append(sys.stdout)
        self._buf = StringIO()
        # Python 3's StringIO objects don't support setting encoding or errors
        # directly and they're already set to None.  So if the attributes
        # already exist, skip adding them.
        if (not hasattr(self._buf, 'encoding') and
                hasattr(sys.stdout, 'encoding')):
            self._buf.encoding = sys.stdout.encoding
        if (not hasattr(self._buf, 'errors') and
                hasattr(sys.stdout, 'errors')):
            self._buf.errors = sys.stdout.errors
        sys.stdout = self._buf

    def restore_stdout(self):
        while self.stdout:
            sys.stdout = self.stdout.pop()
        lg.debug('stdout restored %s', sys.stdout)

    def capture_logging(self):
        lg.debug('call setup_log_handler in EntryRunner')
        setup_log_handler(config.log_handler)

    def restore_logging(self):
        pass

    def _get_buffer(self):
        if self._buf is not None:
            return self._buf.getvalue()

    def _get_logging(self):
        l = map(safe_str, config.log_handler.buffer)
        config.log_handler.truncate()
        return l

    # _prompt_symbol = '➔'
    # _prompt_symbol = '➤'
    _prompt_symbol = '→'

    def log_state(self):
        entry = self.entry
        state = self.state
        status = get_state_status(state)
        full_name = '{}.{}'.format(self.module_runner.module.__name__, entry.__name__)

        start_line = '{} {}... {}'.format(
            color.dye('blue', self._prompt_symbol),
            color.dye('blue', full_name),
            color.dye(STATUS_COLORS[status], status))

        print start_line
        if status == 'FAILED':
            # print hr('=')
            # print hr('-')
            print hr('=')
            print state['traceback']
            if not config.nocapture and state['captured_stdout']:
                print ln('>> begin captured stdout <<')
                print state['captured_stdout']
                print ln('>> end captured stdout <<')
            if not config.nologcapture and state['captured_logging']:
                print ln('>> begin captured logging <<')
                print '\n'.join(state['captured_logging'])
                print ln('>> end captured logging <<')
            print hr('-')
            print ''
        else:
            pass


def log_summary(runners):
    summary = {'UNMET': 0, 'OK': 0, 'FAILED': 0, 'total': 0}
    for runner in runners:
        lg.debug('runner %s', runner)
        if not runner.entries:
            continue
        for entry, state in runner.states.iteritems():
            status = get_state_status(state)
            summary[status] += 1
            summary['total'] += 1

    colored_statuses = {i: (i if summary[i] == 0 else COLORED_STATUSES[i]) for i in STATUS_NAMES}

    print hr('_')
    print 'Ran {s.total} tests, {c.OK} {s.OK}, {c.FAILED} {s.FAILED}, {c.UNMET} {s.UNMET}'.format(
        s=ObjectDict(summary),
        c=ObjectDict(colored_statuses))


def get_state():
    return {
        'unmet': False,
        'executed': False,
        'ok': False,
        'return_value': None,
        'traceback': None,
        'captured_stdout': None,
        'captured_logging': None,
    }


def get_state_status(state):
    if state['unmet']:
        status = 'UNMET'
    elif state['ok']:
        status = 'OK'
    else:
        status = 'FAILED'
    return status


def set_state(state, key, value):
    state[key] = value


def is_failed(state):
    return state['executed'] and not state['ok']


def should_unmet(deps, states):
    unmet = False
    # if any deps not OK or UNMET, should unmet
    for dep in deps:
        state = states[dep]
        if is_failed(state) or state['unmet']:
            unmet = True
    return unmet


def should_pending(deps, states):
    """First a entry is NOT UNMET, then consider if should pending,
    in other words, the prerequisite of PENDING is NOT UNMET
    """
    pending = False
    # if any deps not executed, should pending
    for dep in deps:
        state = states[dep]
        if not state['executed']:
            pending = True
    return pending


def traverse_entry_dependencies(entry, entries_dict, childs=None):
    if childs is None:
        childs = [entry]
    else:
        if entry not in childs:
            childs.append(entry)

    deps = []
    for i in entry.dependencies:
        dep = entries_dict[i['name']]
        #print 'entry', entry, 'dep', dep, 'childs', childs
        if dep in childs:
            raise ValueError(
                'recursive dependency detected: {} depend on {}'.format(entry.__name__, dep.__name__))
        deps.append(dep)
        merge_list(deps, traverse_entry_dependencies(dep, entries_dict, list(childs)))
    return deps


def depend_on(dep_name, with_return=False):
    def decorator_func(f):
        if not hasattr(f, 'dependencies'):
            f.dependencies = []
        # Avoid dep on self
        if dep_name == f.__name__:
            raise ValueError('Depend on self is not allowed')
        # Avoid duplicate dep
        if dep_name in [i['name'] for i in f.dependencies]:
            raise ValueError('Depend on one thing twice is not allowed')
        f.dependencies.append(
            {
                'name': dep_name,
                'with_return': with_return
            }
        )
        return f

    return decorator_func


def with_setup(setup_func=None, teardown_func=None):
    pass


_ignore_dirs = ['.git']


_test_file_pattern = re.compile(r'^(.+_test|test_.+)\.py$')


def walk_dir(dirpath, filepaths):
    for root, dirs, files in os.walk(dirpath):
        for name in files:
            if _test_file_pattern.match(name):
                filepath = os.path.join(root, name)
                filepaths.append(filepath)

        for dir in dirs:
            if dir in _ignore_dirs:
                lg.debug('ignore %s at %s', dir, root)
                dirs.remove(dir)


def define_config(parser):
    config.register_parser(parser)

    config.define('nocapture', 'args')
    parser.add_argument('-s', '--nocapture', action='store_true', help="Don't capture stdout (any stdout output will be printed immediately)")

    config.define('nologcapture', 'args')
    parser.add_argument('--nologcapture', action='store_true', help="Don't capture logging")

    config.define('dry', 'args')
    parser.add_argument('--dry', action='store_true', help="Dry run, only show matched files")

    config.define('debug', 'args')
    parser.add_argument('--debug', action='store_true', help="Set logging level to debug for deptest logger")

    config.define('paths', 'args')
    parser.add_argument('paths', metavar="PATH", type=str, help="files or dirs to scan", nargs='*', default='.')

    config.define('log_handler', None)


def main():
    # the `formatter_class` can make description & epilog show multiline
    parser = argparse.ArgumentParser(description="", epilog="", formatter_class=argparse.RawDescriptionHelpFormatter)

    define_config(parser)

    config.parse_args()

    if config.debug:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO

    set_logger('deptest', level=logging_level, propagate=0,
               fmt='[%(levelname)s %(module)s:%(lineno)d] %(message)s')

    lg.debug('config: %s', config)

    if not config.nologcapture:
        logformat = '%(name)s: %(levelname)s: %(message)s'
        config.log_handler = MyMemoryHandler(logformat)

        lg.debug('call setup_log_handler in global')
        setup_log_handler(config.log_handler)

    filepaths = []

    for path in config.paths:
        if os.path.isdir(path):
            walk_dir(path, filepaths)
        else:
            filepaths.append(path)

    if config.dry:
        for i in filepaths:
            print i
        return

    # Add cwd path to sys.path
    sys.path.insert(0, os.getcwd())

    runners = []
    for filepath in filepaths:
        runner = ModuleRunner(filepath)
        runners.append(runner)
        runner.dispatch()

    log_summary(runners)



# coding: utf-8

"""Util functions for deptest project, unlike tools, it's only for internal use
"""

import re


LINE_WIDTH = 70

# in PY3, its str
unicode_type = unicode

_TO_UNICODE_TYPES = (unicode_type, type(None))


def to_unicode(value):
    """Converts a string argument to a unicode string.

    If the argument is already a unicode string or None, it is returned
    unchanged.  Otherwise it must be a byte string and is decoded as utf8.
    """
    if isinstance(value, _TO_UNICODE_TYPES):
        return value
    if not isinstance(value, bytes):
        raise TypeError(
            "Expected bytes, unicode, or None; got %r" % type(value)
        )
    return value.decode("utf-8")


def ln(label, char='-'):
    """Draw a divider, with label in the middle.

    >>> ln('hello there')
    '---------------------------- hello there -----------------------------'
    """
    label_len = len(label) + 2
    chunk = (LINE_WIDTH - label_len) // 2
    out = '%s %s %s' % (char * chunk, label, char * chunk)
    pad = LINE_WIDTH - len(out)
    if pad > 0:
        out = out + (char * pad)
    return out


def hr(char='-'):
    return LINE_WIDTH * char


def anyp(predicate, iterable):
    for item in iterable:
        if predicate(item):
            return True
    return False


def safe_str(val, encoding='utf-8'):
    try:
        return str(val)
    except UnicodeEncodeError:
        if isinstance(val, Exception):
            return ' '.join([safe_str(arg, encoding)
                             for arg in val])
        return unicode(val).encode(encoding)


def merge_list(a, b):
    for i in b:
        if i not in a:
            a.append(i)


class ObjectDict(dict):
    """
    retrieve value of dict in dot style
    """
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError('Has no attribute %s' % key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError:
            raise AttributeError(key)

    def __str__(self):
        return '<ObjectDict %s >' % dict(self)


def camel_to_underscore(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
